if SCrafts return end

rawset(_G, "SCrafts", true)

if not C2SBarEdits
	print("\x82".."NOTICE: ".."\x80".."S-Crafts are dependent on C2SBarEdits, which must be loaded before this mod in order to be utilized.")
	return
end

------		[READ ME]		------
/*
S-Crafts are special, powerful skills from the Kiseki/Trails game series, the equivalent of super attacks,
and have very special mechanics in order to be made properly.

S-Crafts should use the following flags:

		cost = 0,					--We have to circumvent standard skill costs in order to read the user's current CP amount on use.
		forcetp = true,				--
		notpgain = true,			--
		scraft = true,				--This flag confirms the skill as being an S-Craft for damage and cost handling.
		nometronome = true,			--
		norandom = true,			--

Optional Flags:

		strengthoverride = true, 	--This allows TP skills to use STR instead of MAG for damage calculations.

S-Crafts should have the following nocast function:

		nocast = 	function(targets, n, mo)
				if not charStats[mo.stats].tpbar or not charStats[mo.stats].cpuse
					BTL_logMessage(targets[1].battlen, "You can't use this skill.")
					return true
				end
				if charStats[mo.stats].tpspuse
					if mo.tp < 100
						BTL_logMessage(targets[1].battlen, "Insufficient Craft Points.")
						return true
					end
				else
					if mo.sp < 100
						BTL_logMessage(targets[1].battlen, "Insufficient Craft Points.")
						return true
					end
				end
			end,

S-Crafts require 100 TP to cast, but will zero-out our current TP upon use. The more TP we have when we cast an
S-Craft, the stronger its effect will be. It's also a Kiseki standard for some S-Crafts to have additional effects
when cast at 200 CP (CP, Craft Points, is what we're using TP as a stand-in for.)

-From this point on, we'll refer to TP as being CP unless directly meaning base, unmodified TP behavior-

This addon gives an alternate behavior to TP to make it work as faithfully as possible to the Kiseki CP system.
Characters who use CP have the following TP behavior changes:
	*Fixed 200 CP capacity
	*CP is completely depleted upon death
	*CP is retained between fights instead of resetting at the end of battle

To make a character use the CP system instead of standard TP, they require the following charStats flag:
		cpuse = true,

It should also be noted that defining a character's S-Crafts in their charStats will do nothing if they are not
using the CP system, though the character is not required to have SP functionality (tpspuse charStats flag is optional).
*/

local function addSkill (mo, skill)
			
	local check = false
					
	if #mo.skills
		for s = 1,#mo.skills
		local x = mo.skills[s]
			if skill == x
				check = true
			end
		end
	end
										
	if #mo.passiveskills
		for s = 1,#mo.passiveskills
		local x = mo.passiveskills[s]
			if skill == x
				check = true
				else
			end
		end
	end
									
	if not check
		if attackDefs[skill].type & ATK_PASSIVE
			mo.passiveskills[#mo.passiveskills+1] = skill
		else	
			mo.skills[#mo.skills+1] = skill
		end
	end
end

local function removeSkill (mo, skill)
	
	local storenum = 0
				
	if #mo.skills
		for i = 1,#mo.skills
		local x = mo.skills[i]
			if skill == x
				storenum = i
			end
		end
	end
				
	if #mo.passiveskills
		for i = 1,#mo.passiveskills
			local x = mo.passiveskills[i]
			if skill == x
				storenum = i
			end
		end
	end
				
	if storenum
		if attackDefs[skill].type & ATK_PASSIVE
			table.remove(mo.passiveskills,storenum)
		else	
			table.remove(mo.skills,storenum)
		end
	end
end

-- Main S-Craft function --

local S_Craft = BTL_splitSkills	--S Crafts
BTL_splitSkills = function(v)
	if v and v.valid and v.stats and charStats[v.stats] and charStats[v.stats].persona and not v.enemy
		if charStats[v.stats].s_craft and charStats[v.stats].cpuse and v.level
			if server.gamemode == GM_VOIDRUN
				return
			end
			
			local scraft = 0
			local scraft2 = 0	--up to two S-Craft line slots. Kiseki usually gets 3, but we're already pushing it with "more skill slots" as it is.
			for k,j in ipairs(charStats[v.stats].s_craft or {})
				if j

					if v.level >= j[2]
						-- check skill flags
						if not attackDefs[j[1]].scraft		--Must work with the S Craft triggerdamage passive.
						or not attackDefs[j[1]].forcetp		--Not super important, but I'll enforce this for resource communication.
						or not attackDefs[j[1]].notpgain	--I'll lift this when C2S makes TP gain more open source.
						or not attackDefs[j[1]].nometronome	--Probably a good idea not to let Metronome pull these.
						or not attackDefs[j[1]].norandom	--Ditto above.
						or not attackDefs[j[1]].norogue		--It'd be fine if not for Emerl and Kirby.
						or not attackDefs[j[1]].nocast		--I can't check the exact function here but you better have it anyway.
						--or not attackDefs[j[1]].noenemycast	--We should be able to handle enemies using S-crafts safely now.
							print("\x82".."Warning: ".."\x80".."Loaded skill is not a valid S-Craft. Please be sure to only put S-Craft flagged skills in the S-Craft slot and that they meet the required criteria.")
							playSound(v.battlen, sfx_not)
							break
						end
						-- add skill
						if j[3] >= 2
							scraft2 = j[1]
						else
							scraft = j[1]
						end
					end
				end
			end
			if scraft then v.skills[#v.skills+1] = scraft end
			if scraft2 then v.skills[#v.skills+1] = scraft2 end

			-- autosort skills
			table.sort(v.skills, attackSortFunc)
			
			if v.passiveskills	--Everyone using S Crafts needs this passive.
				table.insert(v.passiveskills, "scraft_boost")
			else
				table.insert(v.skills, "scraft_boost")
			end
		end
	end
	S_Craft(v)
end

-- Enemy S-Craft function --

SRB2P_addHook("NextTurn", function(mo)
	if mo and mo.valid and mo.enemy and enemyList[mo.enemy] and mo.tpbar
	and (mo.cpuse or enemyList[mo.enemy].cpuse) and (mo.s_craft or enemyList[mo.enemy].s_craft)
		local cp = 0
		if (mo.tpspuse or enemyList[mo.enemy].tpspuse) and mo.maxtp == 100		--We have 200 CP capacity, but lose it all on death.
			mo.tp = 10
			mo.maxtp = 200
			dprint("Enemy "..mo.name.." TPSP confirmed, set to 100 CP and 200 CP capacity.")
		elseif mo.maxsp == 100
			mo.sp = 10
			mo.maxsp = 200
			dprint("Enemy "..mo.name.." TP confirmed, set to 100 CP and 200 CP capacity.")
		end
		if (mo.tpspuse or enemyList[mo.enemy].tpspuse)
			cp = mo.tp
		else
			cp = mo.sp
		end
		if cp >= 100 --they can and will ignore nocasts so we have to do this check.
			if server.gamemode == GM_VOIDRUN	--redundant but still wouldn't be fair if players can't when enemies can
				return
			end
			dprint("Enemy "..mo.name.." has "..cp.." CP, adding S-Crafts.")
			local scraft = 0
			local scraft2 = 0	--up to two S-Craft line slots. Kiseki usually gets 3, but we're already pushing it with "more skill slots" as it is.
			for k,j in ipairs((mo.s_craft or enemyList[mo.enemy].s_craft) or {})
				if j
					local level = 1
					if mo.level and mo.level > 1 then level = mo.level end	--in case some one makes an enemy with no level, if that's possible.
					if level >= j[2]
						-- check skill flags
						if not attackDefs[j[1]].scraft		--Must work with the S Craft triggerdamage passive.
						or not attackDefs[j[1]].forcetp		--Not super important, but I'll enforce this for resource communication.
						or not attackDefs[j[1]].notpgain	--I'll lift this when C2S makes TP gain more open source.
						or not attackDefs[j[1]].nometronome	--Probably a good idea not to let Metronome pull these.
						or not attackDefs[j[1]].norandom	--Ditto above.
						or not attackDefs[j[1]].norogue		--It'd be fine if not for Emerl and Kirby.
						or not attackDefs[j[1]].nocast		--I can't check the exact function here but you better have it anyway.
						--or not attackDefs[j[1]].noenemycast	--We should be able to handle enemies using S-crafts safely now.
							print("\x82".."Warning: ".."\x80".."Loaded skill is not a valid S-Craft. Please be sure to only put S-Craft flagged skills in the S-Craft slot and that they meet the required criteria.")
							playSound(mo.battlen, sfx_not)
							break
						end
						-- add skill
						if j[3] >= 2
							scraft2 = j[1]
						else
							scraft = j[1]
						end
					end
				end
			end
			if scraft then addSkill(mo, scraft) end
			if scraft2 then addSkill(mo, scraft2) end

			-- autosort skills
			table.sort(mo.skills, attackSortFunc)
			addSkill(mo, "scraft_boost")	--Everyone using S Crafts needs this passive.
		else	--shouldn't be casting S-Crafts if the requirements are not properly met
			dprint("Enemy "..mo.name.." has "..cp.." CP, removing S-Crafts.")
			if #mo.skills
				local gone = {}
				for i = 1,#mo.skills
				local x = mo.skills[i]
					if attackDefs[x].scraft
						gone[#gone+1] = i
					end
				end
				for i = 1,#gone
					table.remove(mo.skills,gone[i])
				end
			end
		end
	end
end)


attackDefs["scraft_boost"] = {
		name = "S-Craft Boost",
		type = ATK_PASSIVE,
		desc = "Damage booster for S-Craft skills. \nThe more CP you have, the harder \nS-Crafts will hit!",
		passive = PSV_TRIGGERDAMAGE,
		anim = 	function(mo, target, dmg, crit)
					if mo.attack.scraft
						dprint(mo.name.." attempting S-Craft.")
						if (mo.cpuse or (mo.stats and charStats[mo.stats].cpuse) or (mo.enemy and enemyList[mo.enemy].cpuse))
						and (((mo.tpspuse or (mo.stats and charStats[mo.stats].tpspuse) or (mo.enemy and enemyList[mo.enemy].tpspuse)) and mo.tp) or mo.sp)
							dprint(mo.name.." Has CP, checking quantity.")
							local warn = false
							if (mo.tpspuse or (mo.stats and charStats[mo.stats].tpspuse) or (mo.enemy and enemyList[mo.enemy].tpspuse))
								dprint(mo.name.." Has "..mo.tp.." CP.")
								if mo.tp >= 200
									dmg = $*25/10
								elseif mo.tp < 100
									dmg = 0
									warn = true
								else
									dmg = $*(mo.tp)/100
								end
								--mo.tp = 0
							else
								dprint(mo.name.." Has "..mo.sp.." CP.")
								if mo.sp >= 200
									dmg = $*25/10
								elseif mo.sp < 100
									dmg = 0
									warn = true
								else
									dmg = $*(mo.sp)/100
								end
								--mo.sp = 0
							end
							if warn
								BTL_logMessage(mo.battlen, "Invalid S-craft use! Damage set to 0!")
								print("\x82".."Warning: ".."\x80".."Invalid S-Craft use detected! Please be sure to use the provided nocast funtion in your S-Craft's attakdefs to prevent misuse.")
								playSound(mo.battlen, sfx_not)
								mo.setonemore = false
							end
						end
						if not (mo.cpuse or (mo.stats and charStats[mo.stats].cpuse) or (mo.enemy and enemyList[mo.enemy].cpuse))
							BTL_logMessage(mo.battlen, "Invalid S-craft use! Damage set to 0!")
							print("\x82".."Warning: ".."\x80".."Attempted use of S-Craft with an invalid character type detected! Please be sure to have tpbar and cpuse set to true in your charStats.")
							playSound(mo.battlen, sfx_not)
							mo.sp = 0
							dmg = 0
							mo.setonemore = false
						end
						return dmg
					end
				end,
	}

-- CP capacity and death penalty --

	
local function craftpoints(mo)
	if not mo or not mo.valid or not (mo.tpbar or (mo.stats and charStats[mo.stats] and charStats[mo.stats].tpbar)) return end
	if server.gamemode == GM_VOIDRUN return end --not in void run please
	if not mo.cpuse and not (mo.stats and charStats[mo.stats] and charStats[mo.stats].cpuse) return end	--not a craftpoints user
	if not mo.cpuse then mo.cpuse = true end
	if (mo.stats and charStats[mo.stats].tpspuse) and not mo.tpuse then mo.tpuse = true end
	local btl = server.P_BattleStatus[mo.battlen]
	if btl and not btl.running --only out of battle
		if mo.tpuse and mo.maxtp == 100		--We have 200 CP capacity, but lose it all on death.
			mo.tp = 50
			mo.maxtp = 200
			dprint(mo.name.." TPSP confirmed, set to 100 CP and 200 CP capacity.")
		elseif mo.maxsp == 100
			mo.sp = 50
			mo.maxsp = 200
			dprint(mo.name.." TP confirmed, set to 100 CP and 200 CP capacity.")
		end
	end
	if btl and btl.running 
		if mo.tpspuse
			if not mo.hp and mo.tp then mo.tp = max(0, $-4) end
		else
			if not mo.hp and mo.sp then mo.sp = max(0, $-4) end
		end
		if btl.battlestate == BS_ENDTURN
			if mo.attack and mo.attack.scraft	--In case we miss
				if mo.tpspuse
					if mo.tp then mo.tp = 0 end
				else
					if mo.sp then mo.sp = 0 end
				end
				if mo.enemy	--because I guess better baton pass skips the nextturn hook sometimes?!
					dprint("Enemy "..mo.name.." S-Craft finished, removing S-Crafts.")
					if #mo.skills
						local gone = {}
						for i = 1,#mo.skills
						local x = mo.skills[i]
							if attackDefs[x].scraft
								gone[#gone+1] = i
							end
						end
						for i = 1,#gone
							table.remove(mo.skills,gone[i])
						end
					end
				end
			end
		end
	end
end

addHook("MobjThinker", craftpoints, MT_PFIGHTER)


SRB2P_addHook("EnemyInit", function(mo)
	if (mo.enemy and enemyList[mo.enemy].cpuse) and (mo.tpbaar or (mo.enemy and enemyList[mo.enemy].tpbar)) and not mo.cpuse
		mo.cpuse = true
	end
	if mo.cpuse == true
		if (mo.tpspuse or enemyList[mo.enemy].tpspuse) and mo.maxtp == 100		--We have 200 CP capacity, but lose it all on death.
			mo.tp = 50
			mo.maxtp = 200
			dprint("Enemy "..mo.name.." TPSP confirmed, set to 100 CP and 200 CP capacity.")
		elseif mo.maxsp == 100
			mo.sp = 50
			mo.maxsp = 200
			dprint("Enemy "..mo.name.." TP confirmed, set to 100 CP and 200 CP capacity.")
		end
	end
	if mo.enemy == "clonechar" or mo.enemy == "clonechar-hard"
		if mo.clonedfrom and (mo.clonedfrom.cpuse or charStats[mo.clonedfrom.stats].cpuse) and not mo.cpuse
			local charDefs = charStats[mo.clonedfrom.stats]
			mo.tp = mo.clonedfrom.tp
			mo.maxtp = mo.clonedfrom.maxtp
			mo.cpuse = true
			if charStats[mo.clonedfrom.stats].scraft and not mo.scraft
				mo.s_craft = charDefs.scraft
			end
		end
	end
end)


-- Bar Edits TP passives rewrite --

	
attackDefs["red_tpgauge"] = {
	name = "TP - Set up old SP values",
	type = ATK_PASSIVE,
	target = TGT_ALLALLIES,
	desc = "haha, perhaps",
	implicit = "red_tpgauge.pt2",
	passive = PSV_STARTBATTLE,
	anim = function(mo, targets, hittargets, timer)
			if server.oldtp return end
			if mo.cpuse or charStats[mo.stats].cpuse or (mo.enemy and enemyList[mo.enemy].cpuse)
				dprint("Character uses CP, no corrections needed.")
				return
			end
			
			
			mo.oldrealsp = mo.realsp
			mo.oldmaxsp = mo.maxsp
			if mo.tpspuse -- failsafe
				if mo.tp < mo.maxtp/10
					mo.tp = mo.maxtp/10
					dprint("TP is below 10% of max! Setting it to the correct value...")
				end
			else
				if mo.sp < mo.maxsp/10 -- failsafe
					mo.sp = mo.maxsp/10
					dprint("TP is below 10% of max! Setting it to the correct value...")
				end
			end
		end,
	}
	
attackDefs["red_tpgauge.pt2"] = {
	name = "TP - Reset any changed values",
	type = ATK_PASSIVE,
	target = TGT_ALLALLIES,
	desc = "toilet paper",
	passive = PSV_ENDBATTLE,
	anim = function(mo, targets, hittargets, timer)
			if server.oldtp return end
			if mo.cpuse or charStats[mo.stats].cpuse	--enemies don't survive
				dprint("Character uses the CP system, don't reset.")
				if mo.tpspuse
					if not mo.tp	--Bar Edits uses a dungeon thinker hook that forces us back to 10% capacity.
						mo.tp = 1
						dprint("Setting CP to 1 to avoid force reset.")
					end
					mo.teatp = mo.tp
				else
					if not mo.sp	--Bar Edits uses a dungeon thinker hook that forces us back to 10% capacity.
						mo.sp = 1
						dprint("Setting CP to 1 to avoid force reset.")
					end
					mo.teatp = mo.sp
				end
				return
			end
			if mo.tpspuse
				mo.tp = mo.maxtp/10
			else
				mo.sp = mo.maxsp/10
				mo.realsp = mo.oldrealsp
				mo.maxsp = mo.oldmaxsp
			end
			mo.teatp = nil
			dprint("Resetting TP...")
		end,
}

-- Bar Edits TP items rewrite --


itemDefs["diettea"] = {
		name = "Diet Tea",
		desc = "Slight TP recovery to an ally.\nStart next battle at 15% TP\nwhen used outside of battle.",
		attack = "heal_tp_0",
		cost = 900,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+5)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+5)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp*15/100
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp*15/100
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 3,
	}

itemDefs["lucrativetea"] = {
		name = "Lucrative Tea",
		desc = "Moderate TP recovery to an ally.\nStart next battle at 25% TP\nwhen used outside of battle.",
		attack = "heal_tp_1",
		cost = 900,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+10)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+10)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp/4
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp/4
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 3,
	}

itemDefs["volatiletea"] = {
		name = "Volatile Tea",
		desc = "High TP recovery to an ally.\nStart next battle at 50% TP\nwhen used outside of battle.",
		attack = "heal_tp_2",
		cost = 1750,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+20)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+20)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp/2
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp/2
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 4,
	}

itemDefs["explosivetea"] = {
		name = "Explosive Tea",
		desc = "Full TP recovery to an ally.\nStart next battle at 75% TP\nwhen used outside of battle.",
		attack = "heal_tp_3",
		cost = 4000,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+40)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+40)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp*3/4
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp*3/4
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 5,
	}














